package com.tzld.piaoquan.recommend.server.service.flowpool;

import com.tzld.piaoquan.recommend.server.model.TripleConsumer;
import com.tzld.piaoquan.recommend.server.model.Video;
import com.tzld.piaoquan.recommend.server.common.ThreadPoolFactory;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.collections4.MapUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.data.redis.core.RedisTemplate;
import org.springframework.stereotype.Service;

import java.util.*;
import java.util.concurrent.ExecutorService;
import java.util.stream.Collectors;

import static com.tzld.piaoquan.recommend.server.service.flowpool.FlowPoolConstants.*;

/**
 * @author dyp
 */
@Service
@Slf4j
public class FlowPoolService {
    @Autowired
    @Qualifier("redisTemplate")
    private RedisTemplate<String, String> redisTemplate;
    @Autowired
    private FlowPoolConfigService flowPoolConfigService;

    private final String localDistributeCountFormat = "flow:pool:local:distribute:count:%s:%s";

    public final String valueFormat = "%s-%s";

    private ExecutorService pool = ThreadPoolFactory.defaultPool();

    public void updateDistributeCountWithLevel(List<Video> videos) {
        if (CollectionUtils.isEmpty(videos)) {
            return;
        }
        Map<Long, String> removeMap = updateDistributeCount(videos);

        asyncDelDistributeCountWithLevel(removeMap);

    }

    private void asyncDelDistributeCountWithLevel(Map<Long, String> videoFlowPoolMap) {
        asyncDelDistributeCountV2(videoFlowPoolMap, (appType, level, values) -> {
            redisTemplate.opsForSet().remove(String.format(KEY_WITH_LEVEL_FORMAT_V2, appType, level), values);
        });
    }

    private Map<Long, String> updateDistributeCount(List<Video> videos) {
        // TODO 异步更新
        Map<Long, String> removeMap = new HashMap<>();
        videos.stream().forEach(v -> {
            String key = String.format(localDistributeCountFormat, v.getVideoId(), v.getFlowPool());
            Long count = redisTemplate.opsForValue().decrement(key);
            if (count <= 0) {
                removeMap.put(v.getVideoId(), v.getFlowPool());
            }
        });
        return removeMap;
    }

    private void asyncDelDistributeCountV2(Map<Long, String> videoFlowPoolMap,
                                           TripleConsumer<Integer, String, String[]> flowPoolRemoveConsumer) {
        if (MapUtils.isEmpty(videoFlowPoolMap)) {
            return;
        }
        pool.execute(() -> {
            List<String> keys = videoFlowPoolMap.entrySet().stream()
                    .map(v -> String.format(localDistributeCountFormat, v.getKey(), v.getValue()))
                    .collect(Collectors.toList());
            redisTemplate.delete(keys);

            String[] values = new String[videoFlowPoolMap.size()];
            int i = 0;
            for (Map.Entry v : videoFlowPoolMap.entrySet()) {
                values[i++] = String.format(valueFormat, v.getKey(), v.getValue());
            }

            Map<String, Double> levelWeight = flowPoolConfigService.getLevelWeight4FlowPoolWithLevel();
            for (String level : levelWeight.keySet()) {
                flowPoolRemoveConsumer.accept(0, level, values);
            }
        });
    }
}
